#include "cmdqbf.h"
#include "cmdqbf_protos.h"
#include "depqbf/qdpll.h"
#include "glucose/glucose.h"
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>

cmdqbfv cmdq;
masterv master;
statsv stats;

/******************
 * initialization *
 ******************/

int main(int argc, char **argv)
{
	cmdqbf_init(argc, argv);
	mprintf(("%d: initialized on %s\n", cmdq.rank, cmdq.host));

	if (cmdq.rank == MASTER)
		return cmdqbf_master();
	else
		return cmdqbf_worker();
}

static void cmdqbf_init(int argc, char **argv)
{
	int count;
	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &cmdq.rank);
	MPI_Comm_size(MPI_COMM_WORLD, &cmdq.size);
	MPI_Get_processor_name(cmdq.host, &count);
	cmdqbf_initstrucs(); /* initialize default values of globals */
	/* process commandline arguments, set input/output files */
	cmdqbf_commandline(argc, argv);
	/* setup signals --
	 * TERM checkpoint to checkpoint file, or output if none & exit
	 * HUP ditto
	 */
#if 0
	signal(SIGTERM, cmdqbf_caughtsig);
	signal(SIGHUP, cmdqbf_caughtsig);
#endif /* TODO */
}

static void cmdqbf_initstrucs(void)
{
	cmdq.debug = 0;
	cmdq.is_sat = 1; /* existential only, a SAT formula */

	gettimeofday(&master.tstart, NULL);
	master.tree = NULL;
	master.jlist = NULL;
	master.jlist_tails = NULL;
	master.curjobs = NULL;
	master.lresults = NULL;
	master.interrupted = NULL;
	master.randint = INIT_RANDINT;
	master.prefix = NULL;
	master.input_filename = NULL;
	master.input = stdin;
	master.hist_filename = NULL;
	master.hist = NULL;
	master.prefix = NULL;
	master.size_p = 0;
	master.clauses = NULL;
	master.num_clauses = 0;
	master.num_vars = 0;
	master.finished = 0;
	master.varfreq = NULL;
	master.posvarfreq = NULL;
	master.busy_factor = INIT_BUSYFACTOR;
	master.s_busy_factor = INIT_SBF;
	master.budget_scale = INIT_BUDGETSCALE;
	master.add_fac = INIT_BUDGETADDFAC;
	master.init_maxdec = INIT_MAXDEC;
	master.init_maxbtracks = INIT_MAXBTRACKS;
	master.init_maxsecs = INIT_MAXSECS;	
	master.use_laston = INIT_LASTON;
	master.sat_on_sat = INIT_SATONSAT;
	master.rotate = INIT_ROTATE;
	master.rand_steal = INIT_RANDSTEAL;
	master.rem_jobs = 0; /* current #jobs on workers + in job list */
	master.size_jlist = 0; /* current #jobs in job list */
	master.fintime = 0;
	master.gsecb_warn = 0;

	stats.run_jobs = 0;
	stats.fin_jobs = 0;
	stats.max_maxdec = 0;
	stats.max_maxbtracks = 0;
	stats.max_maxsecs = 0;
	cmdq.cleanstart = 1; /* clean start unless error occurs */
	cmdq.relev_assumps = INIT_RELASSUMPS;
	cmdq.diversify = INIT_DIVERSIFY;
	cmdq.result = NULL;
	cmdq.lresult = 0;
	cmdq.alloc_size_result = 0;
	cmdq.alarmsecs = INIT_ALARMSECS;
}

/* Bad commandline arguments. Complain and die. */
static void bad_args(void)
{ /* TODO should MPI_Abort here in case processses get different arguments */
	if (cmdq.rank == MASTER)
	{
		fprintf(stderr, "Invalid arguments.\n");
		printf(USAGE);
	}
	MPI_Finalize();
	exit(0);
}

#define arglong() (i+1<argc ? atol(argv[i+1]) : -1);
/* if "a l" on commandline, set b=l and require l>=c */
#define get_longarg(a, b, c) \
  if (!strcmp(argv[i], a)) { \
     arg = arglong(); \
     if (arg < c) \
       bad_args(); \
     b = arg; \
     i++; \
     continue; \
  }
/* if "a f" on commandline, set b to f */
#define get_filearg(a,b) \
  if (!strcmp(argv[i], a)) { \
    if (i+1>=argc) \
        bad_args(); \
    b = argv[i+1]; \
    i++; \
    continue; \
  }
/* like get_longarg for floats */
#define get_floatarg(a,b,c) \
  if (!strcmp(argv[i], a)) { \
     if (i+1>=argc) \
        bad_args(); \
     test = sscanf(argv[i+1], "%f", &farg); \
     if (test != 1 || farg<c) \
        bad_args(); \
     b = farg; \
     i++; \
     continue; \
  }
/* if "a" on commandline, set b to c */
#define get_boolarg(a,b,c) \
  if (!strcmp(argv[i], a)) { \
    b = c; \
    continue; \
  }
static void cmdqbf_commandline(int argc, char **argv)
{
	int i;
	float farg;
	long arg;
	int test;

	for (i=1; i<argc; i++)
	{
		get_longarg("--max-dec", master.init_maxdec, 0);
		get_longarg("--max-btracks", master.init_maxbtracks, 0);
		get_floatarg("--max-secs", master.init_maxsecs, 0);
		get_floatarg("--scale", master.budget_scale, 1);
		get_longarg("--addfac", master.add_fac, 0);
		get_floatarg("--busy-factor", master.busy_factor, 1);
		get_floatarg("--split-factor", master.s_busy_factor, 1);
		get_boolarg("--no-laston", master.use_laston, 0);
		get_boolarg("--laston", master.use_laston, 1);
		get_boolarg("--no-rel-assumptions", cmdq.relev_assumps, 0);
		get_boolarg("--rel-assumptions", cmdq.relev_assumps, 1);
		get_boolarg("--no-diversify", cmdq.diversify, 0);
		get_boolarg("--diversify", cmdq.diversify, 1);
		get_boolarg("--sat-on-sat", master.sat_on_sat, 1);
		get_boolarg("--no-sat-on-sat", master.sat_on_sat, 0);
		get_boolarg("--some-rotate", master.rotate, 1);
		get_boolarg("--all-rotate", master.rotate, 2);
		get_boolarg("--no-rotate", master.rotate, 0);
		get_floatarg("--random-steal", master.rand_steal, 0);
		get_boolarg("--random-interrupt", master.randint, 0);
		get_boolarg("--no-random-interrupt", master.randint, -1);
		get_longarg("--alarm", cmdq.alarmsecs, 0);
		get_filearg("--hist", master.hist_filename);
		get_boolarg("--debug", cmdq.debug, 10);

		if (master.input_filename != NULL)
			bad_args();
		master.input_filename = argv[i];
	}
}

static void cmdqbf_initfiles(void)
{
	if (master.input_filename != NULL)
	{
		master.input = fopen(master.input_filename, "r");
		if (master.input == NULL)
		{
			printf("Unable to open input file %s\n", master.input_filename);
			cmdq.cleanstart = 0;
		}
	}
	if (master.hist_filename != NULL)
	{
		master.hist = fopen(master.hist_filename, "w");
		if (master.hist == NULL)
			fprintf(stderr,"Unable to open histograph file %s\n",master.hist_filename);
	}
}

/* master sends the instance to process target */
static void master_sendform(int target)
{
	unsigned long buf[5];
	qpref *qp;
	long *clause;
	int i, j;

	buf[0] = cmdq.cleanstart;
	buf[1] = master.num_clauses;
	buf[2] = master.num_vars;
	buf[3] = master.size_p;
	buf[4] = cmdq.is_sat & master.sat_on_sat;

	mprintf(("M: sending instance to %d (%lu,%lu,%lu,%lu,%lu)\n", target, 
		 buf[0],buf[1],buf[2],buf[3],buf[4]));

	/* send cleanstart flag, number of clauses&variables, length of prefix*/
	MPI_Send(buf, 5, MPI_UNSIGNED_LONG, target, 1, MPI_COMM_WORLD);

	if (cmdq.cleanstart != 1)
		return; /* aborting, don't bother sending */

	/* send prefix */
	for (i=0; i<master.size_p; i++)
	{
		qp = &master.prefix[i];
		MPI_Send(&qp->num_vars,1,MPI_UNSIGNED,target,1,MPI_COMM_WORLD);
		MPI_Send(&qp->quant, 1, MPI_CHAR, target, 1, MPI_COMM_WORLD);
		MPI_Send(qp->vars,qp->num_vars,MPI_LONG,target,1,MPI_COMM_WORLD);
	}

	/* send clauses */
	for (i=0; i<master.num_clauses; i++)
	{
		clause = master.clauses[i];
		for (j=0; clause[j]!=0; j++);
		j++; /* also send the closing 0 */
		MPI_Send(&j, 1, MPI_INT, target, 1, MPI_COMM_WORLD);
		MPI_Send(clause, j, MPI_LONG, target, 1, MPI_COMM_WORLD);
	}
}

/* receive formula and add to depqbf */
static void worker_recvform(void)
{
	unsigned long buf[5];
	long *vars = NULL;
	long *clause = NULL;
	int i, num_clauses, num_vars, size_p, nv;
	char q;

	MPI_Recv(buf, 5, MPI_UNSIGNED_LONG, MASTER, 1, MPI_COMM_WORLD,
		 MPI_STATUS_IGNORE);

	cmdq.cleanstart = buf[0];
	num_clauses = buf[1];
	num_vars = buf[2];
	size_p = buf[3];
	cmdq.is_sat = buf[4];

	mprintf2(("%d: got instance header (%lu,%lu,%lu,%lu,%lu)\n", cmdq.rank,
		  buf[0],buf[1],buf[2],buf[3],buf[4]));
	if (cmdq.cleanstart != 1)
		return; /* aborting, remainder not sent */

	allocvars(cmdq.solver, cmdq.sat_solver, num_vars);

	/* get and add prefix */
	for (i=0; i<size_p; i++)
	{
		MPI_Recv(&nv, 1, MPI_UNSIGNED, MASTER, 1, MPI_COMM_WORLD,
			 MPI_STATUS_IGNORE);
		MPI_Recv(&q, 1, MPI_CHAR, MASTER, 1, MPI_COMM_WORLD,
			 MPI_STATUS_IGNORE);
		vars = malloc(sizeof(long)*nv);
		MPI_Recv(vars, nv, MPI_LONG, MASTER, 1, MPI_COMM_WORLD,
			 MPI_STATUS_IGNORE);
		/* add this quantifier block */
		cmdqbf_add_qblock(cmdq.solver, cmdq.sat_solver, vars, q, nv);
		free(vars);
	}

	/* get and add clauses, note we get the closing 0 in each */
	for (i=0; i<num_clauses; i++)
	{
		MPI_Recv(&nv, 1, MPI_INT, MASTER, 1, MPI_COMM_WORLD,
			 MPI_STATUS_IGNORE);
		clause = malloc(sizeof(long)*nv);
		MPI_Recv(clause, nv, MPI_LONG, MASTER, 1, MPI_COMM_WORLD,
			 MPI_STATUS_IGNORE);
		cmdqbf_add_clause(cmdq.solver, cmdq.sat_solver, clause, nv);
		free(clause);
	}

	if (cmdq.debug > 4 && !cmdq.is_sat)
	{
		fprintf(stderr, "== %d got following formula ==\n", cmdq.rank);
		qdpll_print(cmdq.solver, stderr);
	}
}

/* we've parsed the formula, but the master is all alone. :-(
 * give the formula to depqbf and run sequentially
 */
static int master_sequential_qbf(void)
{
	QDPLLResult res;
	QDPLL *solver = qdpll_create();
	int ires;
	mprintf(("M: I'm the lonely master that has to actually work :-(\n"));

	parser_to_depqbf(solver);

	mprintf(("M: calling qdpll_sat()\n"));
	res = qdpll_sat(solver);

	if (res == QDPLL_RESULT_SAT)
		ires=1,printf("SAT\n");
	else if (res == QDPLL_RESULT_UNSAT)
		ires=-1,printf("UNSAT\n");
	else
		ires=0,printf("UNKNOWN?\n");

	qdpll_delete(solver);
	return ires;
}

/* we've parsed the formula, but the master is all alone. :-(
 * it's an existential formula and we want to use Glucose.
 */
static int master_sequential_sat(void)
{
	int res;
	minisat_lbool sres;
	minisat_solver *sat_solver = minisat_new();
	mprintf(("M: I'm the lonely master that has to actualy work :-(\n"));
	mprintf(("M: using SAT solver on existential formula\n"));

	parser_to_sat(sat_solver);
	mprintf(("M: calling minisat_solve()\n"));
	sres = minisat_solve(sat_solver,0,NULL);

	if (sres == minisat_l_False)
		res=-1,printf("UNSAT\n");
	else if (sres == minisat_l_True)
		res=1,printf("SAT\n");
	else
		res=0,printf("UNKNOWN?\n");
	minisat_delete(sat_solver);
	return res;
}

/* we've parsed the formula, but the master is all alone. :-(
 * use the above two functions as a fallback
 */
static int master_sequential(void)
{
	if (cmdq.is_sat && master.sat_on_sat)
		return master_sequential_sat();
	else
		return master_sequential_qbf();
}

static int cmdqbf_fallback(MPI_Request *incoming, int *wnums, int *waiting)
{
	int res;
	printf("c cmdqbf running in fallback sequential DepQBF mode\n");
	res = master_sequential();
	master_cleanup(incoming, wnums, waiting);
	MPI_Finalize();
	exit(res);
}

/* compare frequency of variable *one and *two */
int lt(const void *vone, const void *vtwo)
{
	const long *one = vone, *two = vtwo;
	long fone = master.varfreq[*one];
	long ftwo = master.varfreq[*two];
	if (fone<ftwo)
		return 1;
	else if (fone==ftwo)
		return 0;
	else
		return -1;
}

/* gives garbage for a<=0 */
unsigned int mylog2(unsigned long a)
{
	unsigned int i = 0;
	while (a>1)
	{
		i++;
		a >>= 1;
	}
	return i;
}

static void master_init_incoming(MPI_Request *incoming)
{
	int i;
	int **resbuf = master.resbuf;
	for (i=0; i<cmdq.size; i++)
	{
		if (i == 0)
		{
			incoming[i] = MPI_REQUEST_NULL;
			resbuf[i][0] = 0;
			continue;
		}
		MPI_Irecv(resbuf[i], master.lresults[i], MPI_INT, i, WORKREQ,
			  MPI_COMM_WORLD, incoming+i);
	}
}

/* are there enough jobs? 1: yes, 0: no, try to split */
static int enough_jobs(void)
{
	return (master.rem_jobs >= cmdq.size*master.busy_factor);
}

/* vn came back with unknown result.  Adjust budget or
 * split as desired, put vn or children in job list
 * if (fewer open jobs than number of MPI worker procs 
 * times busyfactor), split vn into fully balanced tree
 */
/* otherwise, multiply budget by factor */
static void adjust_nodebudget(VNode *vn)
{
	int nvars;

	if (enough_jobs())
	{
		mprintf3(("M: enough jobs, scaling budget\n"));
		if (vn->max_dec > 0)
			vn->max_dec = vn->max_dec*master.budget_scale + 
				      master.add_fac;
		if (vn->max_btracks > 0)
			vn->max_btracks = vn->max_btracks*master.budget_scale + 
					  master.add_fac;
		if (vn->max_secs > 0)
			vn->max_secs = vn->max_secs*master.budget_scale + 
				       master.add_fac;
		add_to_joblist(vn);
		return;
	}
	/* otherwise split */
	nvars = mylog2(cmdq.size*master.s_busy_factor - master.rem_jobs);
	if (nvars<1)
		nvars = 1;
	mprintf3(("M: not enough jobs, splitting on %d vars\n", nvars));
	split_tree(vn, nvars);
}

/* there aren't enough jobs, choose a job in the joblist and call
 * adjust_nodebudget(). for now, unassigned job if any, otherwise first
 * found on random worker.
 * TODO: better would be a highest-budget job?
 * TODO: this sort of thing can fail to split if we hit a tiny leaf somewhere
 */
static void master_split(void)
{
	dlist *dl;
	VNode *vn;
	int i, j, r;

	dl = master.jlist[0];
	if (dl!=NULL)
	{
		vn = dl->data;
		remove_jlist(vn);
		adjust_nodebudget(vn);
		return;
	}

	r = rand()%cmdq.size;
	for (i=0; i<cmdq.size; i++)
	{
		j = (r+i)%cmdq.size; /* don't always try low ids first */
		dl = master.jlist[j];
		if (dl!=NULL)
		{
			vn=dl->data;
			remove_jlist(vn);
			adjust_nodebudget(vn);
			return;
		}
	}
}

/* check if the lits at vn match the relevant assumptions ra (size sizera)
 * returns 1 if ra's lits are a subset of vn (all match)
 * returns -1 if some lit in ra occurs w/ opposite sign in vn (def. non-match)
 * returns 0 if a proper subset of ra's lits appear in vn (partial match)
 */
static int match_rela(VNode *vn, int *ra, int sizera)
{
	int i, j, matched=0;
	for (i=0; i<vn->nlits; i++)
	{
		for (j=0; j<sizera; j++)
		{
			if (vn->lits[i] == ra[j])
			{
				matched++;
				break;
			}
			else if (vn->lits[i] == -1 * ra[j])
				return -1; /* not a match */
		}
		if (matched == sizera) /* no point checking the others */
			break;
	}
	if (matched == sizera) /* matched them all! */
		return 1;
	return 0; /* no mis-matches but not all */
}

/* any node in the subtree w/ lits matching ra[0]...ra[sizera-1] can be assigned
 * result and simplify_tree called.
 * lits not in ra are irrelevant, all lits in ra must be matched
 * suffices to look only at leaves (jlist & curjobs):
 * if an inner node matches then all children will too
 * simplify_tree() goes bottom up, easiest to do same here
 * these just came from wnum, no need to check their current job
*/
static void use_relevant_assumptions(int *ra, int sizera, int result)
{
	VNode *vn;
	dlist *dl;
	int i;

	mprintf3(("M: using relevant assumptions (result %d): ",result));
	for (i=0; i<sizera; i++)
		mprintf3(("%d ", ra[i]));
	mprintf3(("\n"));

	for (i=0; i<cmdq.size; i++)
	{
	use_ra_startagain: /* after simplify_tree() dl&dl->next may be lost */
		for (dl = master.jlist[i]; dl; dl=dl->next)
		{
			vn = dl->data;
			if (vn==NULL)
			{
				printf("%d has a NULL job in jlist? why?\n",i);
				continue;
			}
			if (match_rela(vn, ra, sizera) == 1)
			{
				mprintf3(("M: matched all relevant assumptions, simplifying!\n"));
				vn->result = result;
				simplify_tree(vn);
				goto use_ra_startagain;
			}
		}
	}
	for (i=0; i<cmdq.size; i++)
	{
		vn = master.curjobs[i];
		if (vn==NULL)
			continue;
		if (match_rela(vn, ra, sizera) == 1)
		{
			mprintf3(("M: matched all relevant assumptions in %d's job, simplifying(2)!\n",i));
			vn->result = result;
			simplify_tree(vn);
/*#ifdef HORDEQBF_DEPQBF*/
			master_interrupt(i);
/*#endif*/
		}
	}
}

/* process #wnum just returned 'result' for their current job.
 *  (-1: false  0: unknown  1: true)
 * use this to simplify the tree.
 * if result==0, increase budget/split if desired, adding to joblist as needed
 */
static void master_handle_result(int wnum)
{
	VNode *vn = master.curjobs[wnum];
	int result = master.resbuf[wnum][0];
	int sizera = master.resbuf[wnum][1], ackint=master.resbuf[wnum][2],
	    *ra=master.resbuf[wnum]+3;
	master.curjobs[wnum] = NULL;

	if (result == INITIALJOB)
		return;

	mprintf(("M: got result (%d, sizera:%d, ackint:%d) from %d\n", result,
		 sizera, ackint, wnum));

	if (ackint>0)
		master.interrupted[wnum]--;

	vn->on_worker = 0;

	master.rem_jobs--; /* added back by adjust_nodebudget */
	if (result == 0)
	{
		/* adjust budget, put vn or children in job list */
		if (vn->parent!=&master.dummy)
			adjust_nodebudget(vn);
		else
		{
			mprintf2(("M: this unknown result is no longer required\n"));
			free_rem_tree(vn);
			return;
		}
		return;
	}
	stats.fin_jobs++;
	vn->result = result;
	simplify_tree(vn);
	if (sizera > 0 && sizera != master.lresults[wnum]) 
	{  /* use relevant assumptions to hopefully simplify other nodes */
		mprintf3(("M: these relevant assumptions are non-trivial\n"));
		if (!master.finished) /* if simplify_tree won, don't bother */
			use_relevant_assumptions(ra,sizera,result);
	}
	return;
}

/* bid a fond farewell to target */
/* like send_work but sends negative nlits */
static void send_farewell(int target)
{
	long junk[5] = {-1};
	mprintf(("M: bidding farewell to %d; thanks for all the work!\n", 
		 target));
	MPI_Send(junk, 5, MPI_LONG, target, WORKREQ, MPI_COMM_WORLD);
	return;
}

/* get a bit of work for target. guaranteed some work exists.
 * if possible, something that target has seen recently etc.
 * if not, try the common pool. otherwise steal it from someone.
 */
static VNode *get_work(int target)
{
	dlist *dl;
	int i, j, r;
	

	if (master.rand_steal==0 || (rand()>=(master.rand_steal * (float)RAND_MAX)))
	{ /* we don't want to randomly steal */
		dl = master.jlist[target];
		if (dl != NULL)
		{
			/* if not rotating this one, or at least two jobs in list */
			if (master.rotate == 0 || (master.rotate==1 && !(target&1)) || 
			    master.jlist[target]!=master.jlist_tails[target]) /* yay */
				return dl->data;
			/* want to rotate but only one job, steal one from master */
			if (master.jlist[0] != NULL)
				return master.jlist[0]->data;
			return dl->data; /* okay, just keep our single job for now */
		}

		dl = master.jlist[0];
		if (dl != NULL) /* okay */
			return dl->data;
	} /* not wanting to random steal */
	/* either we want to randomly steal, or we have no choice */
	r = rand()%cmdq.size;
	for (i=0; i<cmdq.size; i++)
	{
		j = (r+i)%cmdq.size; /* don't always try low ids first */
		if (master.jlist[j]!=NULL)
			return master.jlist[j]->data;	
	}
	fprintf(stderr, "Error: in get_work but none exists!\n");
	return NULL; /* uh oh, in trouble */
}

/* send a VNode to target, if possible something target has seen
 * recently.  Also remove it from the job list.
 */
static void send_work(int target)
{
	int *tmp;
	long header[5] = {0};
	VNode *vn = get_work(target);
	unsigned int i;

	header[0] = vn->nlits;
	header[1] = 0;
	header[2] = vn->max_dec;
	header[3] = vn->max_btracks;
	MPI_Send(header, 5, MPI_LONG, target, WORKREQ, MPI_COMM_WORLD);
	if (vn->nlits>0)
		MPI_Send(vn->lits, vn->nlits, MPI_LONG, target, WORKREQ,
			 MPI_COMM_WORLD);
	MPI_Send(&vn->max_secs, 1, MPI_FLOAT, target, WORKREQ,
		 MPI_COMM_WORLD);

	if (master.alloc_size_res[target] < vn->nlits + 3)
	{
		tmp = realloc(master.resbuf[target], sizeof(int)*(vn->nlits+3));
		if (tmp == NULL)
		{
			fprintf(stderr, "M: no memory\n");
			MPI_Abort(MPI_COMM_WORLD, 1);
			exit(0);
		}
		master.resbuf[target] = tmp;
		master.alloc_size_res[target] = vn->nlits+3;
	}
	master.lresults[target] = vn->nlits + 3;

	mprintf2(("M: sending work to %d (laston %d, length %d, header %ld,%ld,%ld,%ld,%ld,%f): ", 
		  target,vn->last_on,vn->nlits,header[0],header[1],header[2],header[3],header[4],vn->max_secs));
	for (i=0; i<vn->nlits; i++)
		mprintf2(("%ld ", vn->lits[i]));
	mprintf2(("\n"));

	remove_jlist(vn);
	vn->on_worker = target;
	vn->last_on = (master.use_laston? target : 0);
	master.curjobs[target] = vn;

	if (master.hist == NULL) /* don't update if not needed? */
		return;

	/* update stats used for histogram */
	stats.run_jobs++;
	if (vn->max_dec > stats.max_maxdec)
		stats.max_maxdec = vn->max_dec;
	if (vn->max_btracks > stats.max_maxbtracks)
		stats.max_maxbtracks = vn->max_btracks;
	if (vn->max_secs > stats.max_maxsecs)
		stats.max_maxsecs = vn->max_secs;
	return;
}

static void recv_work(long *header)
{
	int nlits = header[0];
	long *lits = NULL;
	int *tmp;
	int i;
	float max_sec;
	if (nlits>0)
	{	lits = malloc(sizeof(long)*nlits);
		MPI_Recv(lits, nlits, MPI_LONG, MASTER, WORKREQ, MPI_COMM_WORLD,
		 	 MPI_STATUS_IGNORE);
	}
	reset_solver();
	MPI_Recv(&max_sec, 1, MPI_FLOAT, MASTER, WORKREQ, MPI_COMM_WORLD,
		 MPI_STATUS_IGNORE);

	if (cmdq.alloc_size_result < nlits+3)
	{
		tmp = realloc(cmdq.result, sizeof(int)*(nlits+3));
		if (tmp == NULL)
		{
			fprintf(stderr, "%d: no memory\n", cmdq.rank);
			exit(0); /* TODO: don't die, just don't report relev assumps?*/
		}
		cmdq.result = tmp;
		cmdq.alloc_size_result = nlits+3;
	}
	cmdq.lresult = nlits+3;

	/* not conditional to reset any non-HordeQBF interrupted worker */
	/*if (header[2]>0)*/
	set_budget(B_DECS, header[2], 0);
	/*if (header[3]>0)*/
	set_budget(B_BTRACKS, header[3], 0);
	/*if (max_sec>0)*/
	set_budget(B_SECS, 0, max_sec);
	if (cmdq.is_sat && header[2] == 0 && header[3] == 0 && max_sec == 0)
		minisat_no_budget(cmdq.sat_solver);
	mprintf2(("%d: got work (%ld,%ld,%ld,%ld,%ld,%f): ",cmdq.rank,header[0],
                 header[1],header[2],header[3],header[4],max_sec));
	for (i=0; i<nlits; i++)
		mprintf2(("%ld ", lits[i]));
	mprintf2(("\n"));

	set_assumptions(lits, nlits);
	free(lits);
}

/* check if we should update the histogram file timewise, if so do so */
static void do_histogram(int busyw)
{
	struct timeval cur;
	struct timeval *last = &master.last;
	float sec;
	unsigned long gran = 1000000; /* granularity of histogram in microseconds */
	unsigned long inc;

	gettimeofday(&cur,NULL);
	inc = (cur.tv_sec-last->tv_sec)*1000000 + (cur.tv_usec-last->tv_usec);

	if (inc > gran)
	{
		sec = (float)(cur.tv_sec - master.tstart.tv_sec) +
		      ((float)(cur.tv_usec - master.tstart.tv_usec))/1000000;

		fprintf(master.hist, "%f %d %lu %lu %lu %lu %lu %f %f\n", sec,
			busyw, master.size_jlist, stats.run_jobs,stats.fin_jobs,
			stats.max_maxdec,stats.max_maxbtracks,stats.max_maxsecs,
			master.fintime);
		fflush(master.hist);
		last->tv_sec = cur.tv_sec;
		last->tv_usec = cur.tv_usec;
	}		
}

static void master_printinfo(void)
{
	printf("c %s on %d processes\n", VERSION, cmdq.size);
	printf("c budget options max-btracks %ld max-dec %ld max-secs %.2f\n",
	       master.init_maxbtracks, master.init_maxdec, master.init_maxsecs);   printf("c scaling options scale %.2f addfac %ld busy-factor %.2f split-factor %.2f\n",
	       master.budget_scale, master.add_fac, master.busy_factor, 
	       master.s_busy_factor);
      printf("c %susing laston, %susing relevant assumptions, %sdiversifying, %s rotate\n",
	     (master.use_laston?"":"not "), (cmdq.relev_assumps?"":"not "), 
	     (cmdq.diversify?"":"not "),
	     (master.rotate==0?"none":(master.rotate==1?"some":"all")));
	printf("c using %s on DIMACS SAT input\n", 
	       (master.sat_on_sat?"glucose":"depqbf"));
/*#ifdef HORDEQBF_DEPQBF*/
	printf("c alarm time %d seconds, %susing random interrupts\n",
	      cmdq.alarmsecs,(master.randint==-1?"not ":""));
/*#endif*/
	if (master.input_filename != NULL)
		printf("c input from %s\n", master.input_filename);
	if (master.hist_filename != NULL)
		printf("c histogram to %s\n", master.hist_filename);
}

static void master_init(MPI_Request *incoming) 
{
	qpref *block;
	unsigned int i;

	gettimeofday(&master.last, NULL);
	master_printinfo();

	cmdqbf_initfiles();  /* open files on master */
	if (cmdq.cleanstart) /* don't read input if stopping anyways */
		cmdqbf_getinput();

	if (cmdq.debug > 5)
	{
		printf("M: got following formula:\n");
		print_formula();
	}

	master.curjobs = malloc(sizeof(VNode *) * cmdq.size);
	master.resbuf = malloc(sizeof(int *) * cmdq.size);
	master.lresults = malloc(sizeof(unsigned int) * cmdq.size);
	master.interrupted = malloc(sizeof(int) * cmdq.size);
	master.alloc_size_res = malloc(sizeof(unsigned int) * cmdq.size);
	master.jlist = malloc(sizeof(dlist *)*cmdq.size);
	master.jlist_tails = malloc(sizeof(dlist *)*cmdq.size);
	for (i=0; i<cmdq.size; i++)
	{
		master.lresults[i] = 1;
		master.interrupted[i] = 0;
		master.resbuf[i] = malloc(sizeof(int)*16);
		master.alloc_size_res[i] = 16;
		master.jlist[i] = master.jlist_tails[i] = NULL;
		master.curjobs[i] = NULL;
		if (i != MASTER)
			master_sendform(i);
	}
	/* per mpidepqbf, sort variables in each block according to
	 * number of occurrences
	 */
	for (block=master.prefix,i=0; i<master.size_p; block++,i++)
		qsort(block->vars, block->num_vars, sizeof(long), lt);
	master_inittree();

	master_init_incoming(incoming);
}

static void master_cleanup(MPI_Request *incoming, int *wnums, int *waiting)
{
	int i;
	unsigned int ui;
	free(incoming); free(wnums); free(waiting); free(master.curjobs);
	for (i=0; i<cmdq.size; i++)
		free(master.resbuf[i]);
	free(master.resbuf); free(master.lresults);
	free(master.alloc_size_res); free(master.interrupted);

	free_rem_tree(master.tree); /* free tree & all nodes */

	/* free formula */
	free(master.varfreq); free(master.posvarfreq);
	for (ui = 0; ui<master.num_clauses; ui++)
		free(master.clauses[ui]);
	free(master.clauses);
	for (ui = 0; ui<master.size_p; ui++)
		free(master.prefix[ui].vars);
	free(master.prefix);

	if (master.hist != NULL)
		fclose(master.hist);
}

/******************
 * master process *
 ******************/
static int cmdqbf_master(void)
{
	MPI_Request *incoming = malloc(sizeof(MPI_Request) * cmdq.size);
	int *wnums = calloc(cmdq.size, sizeof(int));
	int *waiting = calloc(cmdq.size, sizeof(int));
	int i, count=0, wnum;
	unsigned int busyw = cmdq.size - 1; /* fake initial jobs */
	unsigned int live_workers = cmdq.size - 1;
	int iters = 0; /* do non-essential things only some iterations */

	master_init(incoming);

	if (cmdq.size < 2)
		return cmdqbf_fallback(incoming, wnums, waiting);

	/* main loop */
	while (busyw>0 || live_workers>0)
	{
		/* check for incoming results */
		MPI_Testsome(cmdq.size,incoming,&count,wnums,
			     MPI_STATUSES_IGNORE);
		busyw -= count;

		if (count > 0)
			mprintf2(("M: got %d results\n", count));
/*#ifdef HORDEQBF_DEPQBF*/
		for (i=0; i<count; i++)
		{
			/* already finished, don't interrupt in handle_result*/
			wnum = wnums[i];
			master.interrupted[wnum]++;
		}
/*#endif*/
		/* simplify tree */
		for (i=0; i<count; i++)
		{
			wnum = wnums[i];
			master_handle_result(wnum);
			waiting[wnum] = 1;
			if (master.randint == wnum)
				master.randint = 0;
		}

		for (i=0; i<count; i++)
		{
			/* no need to hold any longer */
			wnum = wnums[i];
			master.interrupted[wnum]--;
		}

		/* hand out new work */
		for (i=0; i<cmdq.size; i++)
		{
			if (waiting[i]==0)
				continue;
			if ( (master.finished == 1) ||
			     (master.size_jlist == 0 && busyw == 0) )
			{ /* want to stop, don't send more work */
				send_farewell(i);
				waiting[i] = 0;
				live_workers--;
			}
			/* send work to wnum */
			else if (master.size_jlist != 0)
			{
				if (!enough_jobs()) /* maybe simplified some */
					master_split();
				send_work(i);
				MPI_Irecv(master.resbuf[i], master.lresults[i],
					  MPI_INT, i, WORKREQ, MPI_COMM_WORLD,
					  incoming+i);
				busyw++;
				waiting[i] = 0;
			}
			/* have no work to send, interrupt random worker
			 * if we haven't done so yet
			 */
			else if (master.randint == 0)
				master.randint = master_rand_interrupt();
				
		}

		/* non-essential stuff */
		iters++;
		if (!(iters & 0x1f))
		{
			if (master.hist != NULL)
				do_histogram(busyw);
		}
	}

	mprintf(("M: out of main loop\n"));

	if (master.finished && cmdq.cleanstart)
	{
		if (master.tree->result == -1)
			printf("UNSAT\n");
		else if (master.tree->result == 0)
			printf("UNKNOWN\n");
		else if (master.tree->result == 1)
			printf("SAT\n");
		else /* hopefully unreachable */
			printf("Things got broken :-(\n");
	}

	master_cleanup(incoming, wnums, waiting);

	MPI_Finalize();
	return 0;
}

/* set budget parameters, call depqbf, put result in result[0].
 * currently result[0] (-1: false, 0: unknown, 1: true)
 * if using relevant assumptions and the job finished, then:
 *    result[1] gives the length of the relevant assumptions which are in
 *    result[3] ... result[3+result[1]-1] (assuming result[1]>0).
 * otherwise result[1] is zero.
 * result[2] currently unused and always 0.
 * values in result[i] for i>=3+result[1]-1 are undefined
 *
 * caller must ensure result is big enough (at least 3 ints + number of
 * assumptions set)
 */
static void do_work(void)
{
	int *result = cmdq.result;
	int *relev_assumps = NULL, sizera;
	int i;
	mprintf2(("%d: calling solver\n",cmdq.rank));
/*#ifdef HORDEQBF_DEPQBF*/
	if (cmdq.stopflag > 1)
		mpi_hordeworker_init(); /* reset interrupt handler */
#ifdef HORDEQBF_DEBQBF
	if (cmdq.stopflag == 1)
		cmdq.stopflag++;
#endif
	alarm(cmdq.alarmsecs); /* reset alarm */
/*#endif*/
	result[0] = run_solver();
/*#ifdef HORDEQBF_DEPQBF*/
	alarm(0); /* alarm off */
/*#endif*/
	mprintf2(("%d: solver returned (%d)\n", cmdq.rank, result[0]));

	/* get relevant assumptions if possible and we set >0 assumptions */
	if (cmdq.relev_assumps && result[0]!=0 && cmdq.lresult>3 &&!cmdq.is_sat)
	{
		relev_assumps = qdpll_get_relevant_assumptions(cmdq.solver);
		for (sizera=0; relev_assumps[sizera]!=0; sizera++);
		/* do a check that sizera small enough */

		mprintf3(("%d: got relevant assumptions, size %d: ", cmdq.rank,
			  sizera));
		for (i=0; i<sizera; i++)
			mprintf3(("%d ", relev_assumps[i]));
		mprintf3(("\n"));

		result[1] = sizera;
		result[2] = cmdq.stopflag; /* were we interrupted? */
		memcpy(result+3, relev_assumps, sizeof(int)*sizera);
		free(relev_assumps);
	}
	else
	{
		result[1] = 0; result[2] = cmdq.stopflag;
	}
	return;
}

static void worker_init(void)
{
	char *opt_buffer;
	int len;

	cmdq.solver = qdpll_create();
	cmdq.sat_solver = minisat_new();

	cmdq.result = calloc(16, sizeof(int));
	cmdq.lresult = 1;
	cmdq.alloc_size_result = 16;
	cmdq.result[0] = INITIALJOB;

	/* TODO remove this once the bug using relevant assumptions with
	 * default variant of clause learning is fixed
	 */
	if (cmdq.relev_assumps)
	{ /* can use --traditional-qcdcl instead or diversify */ 
		qdpll_configure(cmdq.solver, "--no-lazy-qpup");
	}

	cmdq.stopflag = 3; /* set alarm only in do_work */

	/* diversification taken from HordeQBF */
	/* originally from hordesat-src/solvers/DepQBF.cpp:diversify() */
	if (cmdq.diversify)
	{
		len = snprintf(NULL, 0, "--seed=%d", cmdq.rank);
		len++;
		if (len<21)
			len = 21;
		opt_buffer = malloc(sizeof(char)*len);
#ifdef HORDEQBF_DEPQBF
		sprintf(opt_buffer, "--seed=%d", cmdq.rank);
		qdpll_configure(cmdq.solver, opt_buffer);
#endif
		switch (cmdq.rank % 8)
		{
			case 0:
			default:
				sprintf(opt_buffer, "--long-dist-res");
				qdpll_configure(cmdq.solver, opt_buffer);
				sprintf(opt_buffer, "--dep-man=simple");
				qdpll_configure(cmdq.solver, opt_buffer);
			case 1:
#ifdef HORDEQBF_DEPQBF
				qdpll_randomize_assignment_cache(cmdq.solver);
#endif
				break;
			case 2:
				sprintf(opt_buffer, "--no-qbce-dynamic");
				qdpll_configure(cmdq.solver, opt_buffer);
			case 3:
#ifdef HORDEQBF_DEPQBF
				qdpll_randomize_var_activity_scaling(cmdq.solver);
#endif
				break;
			case 4:
				sprintf(opt_buffer, "--no-qbce-dynamic");
				qdpll_configure(cmdq.solver, opt_buffer);
				sprintf(opt_buffer, "--qbce-inprocessing");
				qdpll_configure(cmdq.solver, opt_buffer);
			case 5:
#ifdef HORDEQBF_DEPQBF
				qdpll_randomize_learned_constraint_removal(cmdq.solver);
#endif
				break;
			case 6:
				sprintf(opt_buffer, "--no-qbce-dynamic");
				qdpll_configure(cmdq.solver, opt_buffer);
				sprintf(opt_buffer, "--qbce-preprocessing");
				qdpll_configure(cmdq.solver, opt_buffer);
			case 7:
#ifdef HORDEQBF_DEPQBF
				qdpll_randomize_restart_schedule(cmdq.solver);
#endif
				break;
		}
		free(opt_buffer);
	} /* diversification */
}

static int cmdqbf_worker(void)
{
	MPI_Request req;
	long header[5]={0};

	worker_init();
	worker_recvform();

	MPI_Recv_init(header, 5, MPI_LONG, MASTER, WORKREQ, MPI_COMM_WORLD,
		      &req);
	MPI_Start(&req);
	MPI_Send(cmdq.result, 1, MPI_INT, MASTER, WORKREQ, MPI_COMM_WORLD);
	while (1)
	{
		mprintf2(("%d: waiting for message from master\n",cmdq.rank));
		MPI_Wait(&req, MPI_STATUS_IGNORE);
		{
			if (header[0]<0)
				break; /* finished */
			recv_work(header); /* also allocs cmdq.result */
			do_work(); /* puts answer in cmdq.result */
			MPI_Start(&req);
			MPI_Send(cmdq.result, cmdq.lresult, MPI_INT, MASTER,
				 WORKREQ, MPI_COMM_WORLD);
		}
	}
	mprintf(("%d: out of main loop\n", cmdq.rank));
	qdpll_delete(cmdq.solver);
	minisat_delete(cmdq.sat_solver);
	MPI_Finalize();
	return 0;
}
