/* cmdqbf.h
 *
 */
#ifndef CMDQBFH
#define CMDQBFH

#include "depqbf/qdpll.h"
#include "glucose/glucose.h"
#include <mpi.h>
#include <sys/time.h>

#define VERSION "cmdqbf 0.00"

#define MASTER 0
#define INITIALJOB -100 /* something not -1,0,1 */

#define USAGE "USAGE: cmdqbf <filename> --max-btracks <int> --max-dec <int> --max-secs <float> --scale <float> --addfac <int> --busy-factor <float> --split-factor <float> --no-laston --laston --no-rel-assumptions --rel-assumptions --no-diversify --diversify --sat-on-sat --no-sat-on-sat --no-rotate --some-rotate --all-rotate --random-steal <float> --random-interrupt --no-random-interrupt --alarm <int> --hist <filename> --debug\n"

#define INIT_BUSYFACTOR 1.25 /* default busy factor for number of jobs */
#define INIT_SBF 2	     /* default busy factor for initial tree split */
#define INIT_BUDGETSCALE 1.4 /* default scaling factor for budget */
#define INIT_BUDGETADDFAC 0  /* default additive scale for budget */
#define INIT_MAXDEC 0        /* initial budget for --max-dec,     0 disables*/
#define INIT_MAXBTRACKS 0    /* initial budget for --max-btracks, 0 disables */
#define INIT_MAXSECS 1       /* initial budget for --max-secs,    0 disables */
				/* note mpidepqbf used 0.1 as starting timeout*/
#define INIT_RELASSUMPS 1    /* should use relevant assumptions by default */
#define INIT_DIVERSIFY 0     /* don't diversify by default */
#define INIT_SATONSAT 0	     /* use DepQBF (not Glucose) on DIMACS by default */
#define INIT_ROTATE 2	     /* all rotate by default */
#define INIT_RANDSTEAL 0     /* don't randomly steal by default */
#define INIT_LASTON 1	     /* use laston by default */
#define INIT_RANDINT 0	     /*use random interrupts to get work, -1 disables*/
#define INIT_ALARMSECS 2     /* interrupt every this-many seconds, to check for
			      * interrupt message from master
			      * used only with HordeQBF version of DepQBF
			      */

typedef struct dlist_v {
	void *data;
	struct dlist_v *prev;
	struct dlist_v *next;
} dlist;

/* A structure containing the state of this process.
 * Each process has one.
 */
typedef struct cmdqbf_v {
	QDPLL *solver;
	minisat_solver *sat_solver;
	/* MPI info */
	int rank;		/* my rank */
	int size;		/* number of MPI processes */
	char host[MPI_MAX_PROCESSOR_NAME];	/* name of this host */
	int *result; /* buffer for outgoing result & relevant assumptions */
	unsigned int lresult; /* length of result */
	unsigned int alloc_size_result; /* result allocated size (>=lresult) */

	int relev_assumps; /* 0: don't use relevant assumptions, 1: do */
	int diversify; /* 1: use hordeqbf-style diversify, 0: don't */
	int cleanstart;
	MPI_Request stopreq; /* for checking interrupts w/ HordeQBF DepQBF */
	int stopflag; /* for HordeQBF DepQBF */
	int debug; /* debugging level on this process */
	int alarmsecs; /* time for alarm() with HordeQBF DepQBF */
	int is_sat; /* existential formula, can use a SAT solver */
} cmdqbfv;

typedef struct {
	long *vars;
	unsigned int num_vars;
	char quant; /* a or e */
} qpref;

typedef struct vnode_s {
        long *lits;
	unsigned int nlits;

	long max_dec;
	long max_btracks;
	float max_secs;

	int result; /* -1: false  0: unknown  1: true */
        char quant;  /* 'e': exists  'a': forall */
	dlist *in_jlist; /* pointer to entry in job list, or NULL */
	int on_worker; /* 0: no worker, otherwise current worker#*/
	int last_on; /* preferred worker (or 0:any), used for locality */

	unsigned int nchildren;     /* size of children */
	unsigned int live_children; /* number of children remaining */
				    /* live children==0 has different meaning
				     * in nodes that have/have not been split*/
        struct vnode_s *parent;
        struct vnode_s **children; /* 2^nchildren many */

	int my_cnum; /* I'm parent->children[my_cnum] */
	int pref;    /* first remaining prefix block is master.prefix[pref] */
	unsigned int offset; /* remaining prefix starts at pref->vars[offset] */
} VNode;

/* A structure containing things needed by the master. */
typedef struct cmdqbf_master_v {
	/* quantifier prefix */
	VNode *tree;
	dlist **jlist; /* job list, one per process to try for locality */
	dlist **jlist_tails; /* tails of job lists */
	VNode **curjobs; /* curjobs[i] is the job currently worked on by i
			  * (or NULL); vn->on_worker gives the relevant i
			  */
	int **resbuf; /* result buffers for the workers */
	unsigned int *lresults; /* length of next result for each worker */
	unsigned int *alloc_size_res; /* allocated sizes for resbuf */
	int *interrupted; /* this worker interrupt how many times this job? */
	int randint; /* for randomly interrupting working worker for split */
	VNode dummy; /* a silly dummy placeholder */
	qpref *prefix;
	unsigned int size_p; /* number of quantifier blocks, #alts + 1 */
	long **clauses; /* clause 0 is clauses[0][0]... until a 0 */
	unsigned int num_clauses;
	unsigned int num_vars;
	int finished; /* 0: not yet, 1: finished */
	float fintime; /* number of seconds until found answer */

	long *varfreq; /* #occurrences for each variable |num_vars|+1 length*/
	long *posvarfreq; /* #positive occurrences for each, same length */

	float budget_scale; /* scaling factor for increasing budget */
	long add_fac; /* additive factor when increasing budget */
	long init_maxdec;
	long init_maxbtracks;
	float init_maxsecs;	
	float busy_factor;   /* split when below busy_factor */
	float s_busy_factor; /* busy factor to target when splitting */
	int use_laston; /* 1: use last_on for locality, 0: ignore it */
	int rotate; /* 0: none 1: some (=all if no laston) 2: all */
	int sat_on_sat; /* use 0: DepQBF, 1: Glucose on DIMACS SAT input */
	float rand_steal; /* workers try to randomly steal another's work with
			   * this probability. only makes sense with use_laston,
			   * otherwise just wastes insignificant time
			   */

	unsigned long rem_jobs; /* #jobs on workers + in job list */
	unsigned long size_jlist; /* #jobs in job list */

	/* files */
	char *input_filename;
	FILE *input;
	char *hist_filename;
	FILE *hist;

	struct timeval tstart;
	struct timeval last;

	/* flags, to consolidate if more are added */
	int gsecb_warn; /* already warned for non-zero seconds budget
			 * when using Glucose?
			 */
} masterv;

typedef struct cmdqbf_stats_v {
	unsigned long run_jobs;
	unsigned long fin_jobs;
	unsigned long max_maxdec;
	unsigned long max_maxbtracks;
	float max_maxsecs;
} statsv;

#define B_BTRACKS 1
#define B_DECS 2
#define B_SECS 4

/* fixed MPI tags */
#define WORKREQ 10
#define STOPFLAG 20

#ifdef DEBUG
#define mprintf(a) printf a
#else
#define mprintf(a)
#endif
#ifdef DEBUG2
#define mprintf2(a) printf a
#else
#define mprintf2(a)
#endif
#ifdef DEBUG3
#define mprintf3(a) printf a
#else
#define mprintf3(a)
#endif
#ifdef DEBUG4
#define mprintf4(a) printf a
#else
#define mprintf4(a)
#endif

#endif /* CMDQBFH */
