/* tree.c
 * code related to dealing with the tree of partial assignments,
 * and joblists of the leaves
 * intended to be indepedent of MPI and DepQBF
 */

#include "cmdqbf.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

extern masterv master;
extern cmdqbfv cmdq;

/* following are mostly intended to move here */
extern void master_interrupt(int); /* this won't move here */
extern unsigned int mylog2(unsigned long);

void freenode(VNode *node)
{
	free(node->children);
	free(node->lits);
	free(node);
	return;
}

/* remove vn from master.jlist, node vn.in_jlist points to the entry */
void remove_jlist(VNode *vn)
{
	dlist *d = vn->in_jlist;
	if (d == NULL)
	{
		fprintf(stderr,"%d: attempting to remove node (last_on %d, on_worker %d) from jlist with NULL in_jlist\n",cmdq.rank,vn->last_on,vn->on_worker);
		return;
	}

	if (d->prev != NULL)
		d->prev->next = d->next;
	else
		master.jlist[vn->last_on] = d->next;
	if (d->next != NULL)
		d->next->prev = d->prev;
	else /* equiv: (master.jlist_tails[vn->laston] == d) */
		master.jlist_tails[vn->last_on] = d->prev;
	free(d);
	vn->in_jlist = NULL;
	master.size_jlist--;
	return;
}

/* remove all nodes in subtree from the job list */
/* free all we can, but nodes on workers must be left
 * in the on_worker list until they return
 */
void free_rem_tree(VNode *vn)
{
	VNode *tmp;
	unsigned int i;

	for (i=0; i<vn->nchildren && vn->live_children>0; i++)
	{
		tmp = vn->children[i];
		if (tmp == NULL)
			continue;
		free_rem_tree(tmp);
		vn->children[i] = NULL;
		vn->live_children--;
	}

	if (vn->in_jlist != NULL)
	{
		master.rem_jobs--;
		remove_jlist(vn);
	}

	if (vn->on_worker == 0)
		freenode(vn);
	else /* remove pointer to parent. needed for simplify_tree() called */
	{
		vn->parent = &master.dummy; /* when this job returns */
/*#ifdef HORDEQBF_DEPQBF*/
		master_interrupt(vn->on_worker); /* stop that worker */
/*#endif*/
	}
}

/* we just got a result at node start.  Propagate this result up
 * the tree as much as possible, freeing memory and removing things
 * from the joblist when possible.
 * Must be careful about nodes that are already on other workers
 * We may know their results and can remove from the tree, but must
 * live in the curjobs array for when their current worker returns.
 */
void simplify_tree(VNode *start)
{
	VNode *parent = start->parent;
	struct timeval cur;
	int result = start->result;
	char quant;

	if (parent == NULL)
	{
		/* we won */
		master.finished = 1;
		mprintf(("M: simplify_tree at root, we won!\n"));
		gettimeofday(&cur,NULL);
		master.fintime=(float)(cur.tv_sec - master.tstart.tv_sec) +
			((float)(cur.tv_usec - master.tstart.tv_usec))/1000000;
/*#ifdef HORDEQBF_DEPQBF*/
		master_interrupt(MASTER);
/*#endif*/
		return;
	}
	else if (parent == &master.dummy)
	{
		mprintf2(("M: simplifying a result no longer needed\n"));
		free_rem_tree(start);
		return;
	}

	quant = parent->quant;
	parent->children[start->my_cnum] = NULL;
	parent->live_children--;

	/* remove all nodes in subtree from joblist */
	/* also free those not on workers */
	free_rem_tree(start);

	/* avoid BUG: if found e.g. univ. counterexample & set parent's result
	 * but later think we checked all children positively
	 */
	if (parent->result != 0) /* don't set result if already set */
		return;

	if (quant == 'e')
	{
		if (result == 1)
		{
			mprintf3(("M: found an existential witness\n"));
			parent->result = 1;
			simplify_tree(parent);
			return;
		}
		else if (parent->live_children == 0) /* was last one */
		{
			mprintf3(("M: tried all existential possibilities\n"));
			parent->result = -1;
			simplify_tree(parent);
			return;
		}
		return; /* removed this choice but more remain */
	}
	else if (quant == 'a')
	{
		if (result == -1)
		{
			mprintf3(("M: found universal counterexample\n"));
			parent->result = -1;
			simplify_tree(parent);
			return;
		}
		else if (parent->live_children == 0) /* was last one */
		{
			mprintf3(("M: tried all universal possibilities\n"));
			parent->result = 1;
			simplify_tree(parent);
			return;
		}
		return; /* verified this choice but more remain */
	}
	fprintf(stderr, "Error: unknown quantifier (%c) in tree\n", quant);
	return;
}

/* add node to master.jlist */
void add_to_joblist(VNode *node)
{
	dlist *d = malloc(sizeof(dlist));
	dlist *jlist;
	d->data = node;
	node->in_jlist = d;
	master.rem_jobs++;
	master.size_jlist++;

	/* add at head if not rotating, or rotating some and this one even */
	if (master.rotate == 0 || (master.rotate == 1 && !(node->last_on & 1)))
	{
		jlist = master.jlist[node->last_on];
		d->prev = NULL;
		d->next = jlist;
		if (jlist != NULL)
			jlist->prev = d;
		else /* empty list, also make d tail */
			master.jlist_tails[node->last_on] = d;
		master.jlist[node->last_on] = d;
	}
	/* shuffle this one; add at tail */
	else
	{
		jlist = master.jlist_tails[node->last_on];
		d->next = NULL;
		d->prev = jlist;
		if (jlist != NULL)
			jlist->next = d;
		else /* empty list, also make d head */
			master.jlist[node->last_on] = d;
		master.jlist_tails[node->last_on] = d;
	}
	return;
}

/* vn is a newly allocated node.  Set the default budget. */
static void set_newnodebudget(VNode *vn)
{
	vn->max_dec = master.init_maxdec;
	vn->max_btracks = master.init_maxbtracks;
	vn->max_secs = master.init_maxsecs;
}

/* split node into evenly nvars more variables if possible, or as many as
 * possible if not. Note that this may cross quantifier blocks.
 * return can generally be ignored; we modify subtree rooted at node.
 * leaves of the new subtree are added to the joblist.
 * node is added to the joblist if it can't be split, unless it's already
 * there.  You should probably take node out of the joblist before calling
 * this (and also set node->in_jlist=NULL).
 */
/* must not split nodes that are already split (i.e. node->children must be 0)*/
VNode *split_tree(VNode *node, int nvars)
{
	VNode *vn;
	int rvars = nvars; /*number of variables remaining to split*/
	int svars; /*number of variables to split in current quantifier block*/
	int pref = node->pref;
	unsigned int i, j;
	int nsplit;
	qpref *q;

	if (master.prefix == NULL ||
	pref>=master.size_p)
	{       /* can't split any more */
		mprintf3(("M: want to split %d variables, but all done\n",
			  nvars));
		if (node->in_jlist == NULL)
		add_to_joblist(node);
		return node;
	}

	q = &master.prefix[pref];

	svars = q->num_vars - node->offset;
	mprintf3(("M: want to split %d variables, can do %d in this block\n",
		  rvars, svars));

	if (svars >= rvars)
	{       /* can do whole split in this quantifier block */
		svars = rvars;
		rvars = 0;
	}
	else
	{       /* split svars in this block, want rvars more in later blocks */
		rvars -= svars;
	}

	/* last split was exactly on a quantifier block boundary */
	if (svars == 0)
	{
		mprintf3(("M: in the svars == 0 case !\n"));
		if (pref+1==master.size_p)
		{
			/* out of luck, we've split every variable */
			if (node->in_jlist == NULL)
				add_to_joblist(node);
			return node;
		}
		/* prepare a new node for the next block, to be split */
		vn = malloc(sizeof(VNode));
		vn->nlits = node->nlits;
		vn->lits = malloc(sizeof(long)*vn->nlits);
		memcpy(vn->lits, node->lits, sizeof(long)*node->nlits);
		vn->result = 0;
		vn->quant = master.prefix[pref+1].quant;
		vn->in_jlist = NULL;
		vn->on_worker = 0;
		vn->last_on = (master.use_laston? node->last_on : 0); /*inherit preferred worker*/
		vn->nchildren = 0;
		vn->live_children = 0;
		vn->parent = node;
		vn->children = NULL;
		vn->pref = pref+1;
		vn->offset = 0;
		set_newnodebudget(vn);
		/* TODO: make sure this is okay */
		if (node->nchildren != 0)
		{
			printf(" !!!!! M: in broken svars==0 bit, hit the nchildren assertion !!!!!\n");
			exit(1);
		}
		mprintf3(("! M: svars==0 quantifier flip !\n"));
		node->quant = vn->quant;
		node->nchildren = 1;
		node->live_children = 1;
		node->children = malloc(sizeof(VNode *));
		node->children[0] = vn;
		vn->my_cnum = 0;

		/* do whole split in next depth */
		split_tree(vn, nvars);
		return node;
	}

	/* otherwise, some to split in this block */

	/* split next svars-many */
	nsplit = 2<<(svars - 1);
	node->nchildren = nsplit;
	node->children = malloc(sizeof(VNode *)*(nsplit));
	node->live_children = nsplit;

	mprintf3(("M: splitting node into %d children\n", node->nchildren));
	for (i=0; i<node->nchildren; i++)
	{
		node->children[i] = malloc(sizeof(VNode));
		vn = node->children[i];
		set_newnodebudget(vn);
		vn->my_cnum = i;
		vn->nlits = node->nlits + svars;
		vn->lits = malloc(sizeof(long)*vn->nlits);
		memcpy(vn->lits, node->lits, sizeof(long)*node->nlits);
		memcpy(vn->lits + node->nlits,
		       master.prefix[node->pref].vars+node->offset,
		       sizeof(long)*svars); /* copy new positive literals */
		/* next,
		 * negate literals in positions corresponding to 1s in
		 * binary expansion of i
		 */
		for (j=0; j<svars; j++)
			if ( (i>>j)&1 )
				vn->lits[node->nlits + j] *= -1;
		vn->result = 0;
		vn->quant = node->quant;
		vn->in_jlist = NULL;
		vn->on_worker = 0;
		vn->last_on = (master.use_laston? node->last_on : 0); /*inherit preferred worker*/
		vn->nchildren = 0;
		vn->live_children = 0;
		vn->parent = node;
		vn->children = NULL;
		vn->pref = node->pref;
		vn->offset = node->offset + svars;
		if (rvars > 0) /* do remaining split */
			split_tree(vn, rvars);
		else /* vn is a leaf in the tree */
			add_to_joblist(vn);
	}

	return node;
}

/* per mpidepqbf, initialize master.tree to a fully-balanced
 * binary tree with size highest power of 2 smaller than
 * (#proc-1) * master.s_busy_factor. if not enough variables,
 * just do what we can.
 * we don't restrict to binary, but each level in the tree
 * is at most one quantifier block.
 */
void master_inittree(void)
{
	unsigned int depth = 0;
	VNode *vn = malloc(sizeof(VNode));

	vn->lits = NULL;
	vn->nlits = 0;
	vn->quant = 'e'; /* free variables are existential */
	if (master.size_p>0)
		vn->quant = master.prefix[0].quant;
	vn->result = 0;
	vn->in_jlist = NULL;
	vn->on_worker = 0;
	vn->last_on = 0;
	vn->nchildren = 0;
	vn->parent = NULL;
	vn->children = NULL;
	vn->live_children = 0;
	vn->pref = 0; /* vn->pref=master.prefix; */
	vn->offset = 0;
	vn->my_cnum = -1;
	set_newnodebudget(vn);

	master.tree = vn;

	if (master.num_vars>0)
		depth = mylog2(cmdq.size*master.s_busy_factor);
	if (depth > master.num_vars)
		depth = master.num_vars;

	mprintf2(("M: initial split on %d variables\n", depth));

	if (depth != 0)
		split_tree(master.tree, depth);
	else
		add_to_joblist(vn);
}
