/* util.c
 * some convenient functions used in multiple places
 * so far just solver-interfacing, perhaps will rename to solver.c?
 */

#include "cmdqbf.h"
#include "depqbf/qdpll.h"
#include <stdlib.h>

extern masterv master;
extern cmdqbfv cmdq;

/* for adding the output of the parser directly to depqbf.
 * used in master_sequential and the pthreads version
 * the MPI version usually uses send_form() / recv_form() instead.
 */
void parser_to_depqbf(QDPLL *solver)
{
	qpref *qp;
	long *clause;
	int i, j;
	qdpll_adjust_vars(solver, master.num_vars);
	/* set prefix */
	for (i=0; i<master.size_p; i++)
	{
		qp = &master.prefix[i];
		qdpll_new_scope(solver, (qp->quant=='e' ? QDPLL_QTYPE_EXISTS :
					 QDPLL_QTYPE_FORALL));
		for (j=0; j<qp->num_vars; j++)
			qdpll_add(solver, qp->vars[j]);
		qdpll_add(solver, 0); /* close scope */
	}       

	/* add clauses */
	for (i=0; i<master.num_clauses; i++)
	{
		clause = master.clauses[i];
		for (j=0; clause[j]!=0; j++)
			qdpll_add(solver, clause[j]);
		qdpll_add(solver, 0); /* end the clause */
	}

	if (cmdq.debug > 5)
	{
		fprintf(stderr, "M: added following formula to solver:\n");
		qdpll_print(solver, stderr);
	}
}

/* for adding the output of the parser directly to Glucose. Used for
 * the fallback single process version on existential input when
 * --sat-on-sat is used.
 */

void parser_to_sat(minisat_solver *sat_solver)
{
	long *clause;
	int i, j, sign, var, lit;

	for (i=0; i<master.num_vars; i++) /*don't freeze in single-thread case*/
	{
		minisat_newVar(sat_solver);
		/* minisat_setFrozen(sat_solver, i, minisat_l_True); */
	}

	/* add clauses */
	for (i=0; i<master.num_clauses; i++)
	{
		clause = master.clauses[i];
		minisat_addClause_begin(sat_solver);
		for (j=0; clause[j]!=0; j++)
		{
			sign = (clause[j]>0 ? 1 : 0);
			var = (sign ? clause[j] : -1 * clause[j]);
			var--;
			lit = minisat_mkLit(var);
			if (!sign)
				lit = minisat_negate(lit);
			minisat_addClause_addLit(sat_solver, lit);
		}
		minisat_addClause_commit(sat_solver);
	}
	return;
}

/* alloc variables in solver */
void allocvars(QDPLL *solver, minisat_solver *sat_solver, int num_vars)
{
	int i;
	if (!cmdq.is_sat) /* using DepQBF */
	{
		qdpll_adjust_vars(cmdq.solver, num_vars);
		return;
	}
	/* using SAT */
	for (i=0; i<num_vars; i++)
	{ /* abusive but works and simple */
		minisat_newVar(sat_solver);
		minisat_setFrozen(sat_solver, i, minisat_l_True);
	}
}

/* omits closing 0 */
void cmdqbf_add_qblock(QDPLL *solver, minisat_solver *sat_solver, long *vars, 
		       char q, int nv)
{
	int j;

	if (cmdq.is_sat) /* not using DepQBF */
	{
		if (q == 'a')
		{
			fprintf(stderr, "%d: universal quantifier but master claimed DIMACS!\n", cmdq.rank);
			/*MPI_Abort(MPI_COMM_WORLD, 1);*/
			exit(1);
		}
		return;
	}

	/* otherwise using DepQBF as normal */
	if (q == 'e')
		qdpll_new_scope(solver, QDPLL_QTYPE_EXISTS);
	else if (q == 'a')
		qdpll_new_scope(solver, QDPLL_QTYPE_FORALL);
	else /* unreachable assuming MPI unbroken */
		printf("%d: unknown quantifier %c!\n", cmdq.rank, q);
	for (j=0; j<nv; j++)
		qdpll_add(solver, vars[j]);
	qdpll_add(solver, 0); /* close scope */
}

/* includes closing 0 */
void cmdqbf_add_clause(QDPLL *solver, minisat_solver *sat_solver, long *clause, 
		       int nv)
{
	int j, var, sign, lit;

	if (!cmdq.is_sat) /* using DepQBF */
	{
		for (j=0; j<nv; j++)
			qdpll_add(solver, clause[j]);
		return;
	}

	/* using Glucose */
	minisat_addClause_begin(sat_solver);
	for (j=0; j<nv; j++)
	{
		sign = (clause[j]>0 ? 1 : 0);
		var = (sign ? clause[j] : -1 * clause[j]);
		if (var == 0)
			break;
		var--;
		lit = minisat_mkLit(var);
		if (!sign)
			lit = minisat_negate(lit);
		minisat_addClause_addLit(sat_solver, lit);
	}
	minisat_addClause_commit(sat_solver);
	return;
}

void reset_solver(void)
{
	if (!cmdq.is_sat) /* using DepQBF */
		qdpll_reset(cmdq.solver);
	else
		minisat_solve_begin(cmdq.sat_solver);
}

void set_budget(int type, const long lbudg, const float fbudg)
{
	char *budget = NULL;
	char *ret;
	int blen;

	switch (type)
	{
		case B_BTRACKS:
			if (cmdq.is_sat) /* using Glucose */
			{
				minisat_set_conf_budget(cmdq.sat_solver, lbudg);
				return;
			}
			/* using DepQBF */
			blen = snprintf(NULL, 0, "--max-btracks=%ld", lbudg);
			budget = malloc(sizeof(char)*(blen+2));
			snprintf(budget, blen+1, "--max-btracks=%ld", lbudg);
			break;
		case B_DECS:
			if (cmdq.is_sat) /* using Glucose */
			{
				minisat_set_dec_budget(cmdq.sat_solver, lbudg);
				return;
			}
			/* using DepQBF */
			blen = snprintf(NULL, 0, "--max-dec=%ld", lbudg);
			budget = malloc(sizeof(char)*(blen+2));
			snprintf(budget, blen+1, "--max-dec=%ld", lbudg);
			break;
		case B_SECS:
			if (cmdq.is_sat) /* using Glucose */
			{
				if (fbudg > 0 && !master.gsecb_warn)
				{
					master.gsecb_warn = 1;
					fprintf(stderr, "Warning: budgeting with seconds not supported in Glucose\n");
				}
				return;
			}
			/* using DepQBF */
			blen = snprintf(NULL, 0, "--max-secs=%f", fbudg);
			budget = malloc(sizeof(char)*(blen+2));
			snprintf(budget, blen+1, "--max-secs=%f", fbudg);
			break;
		default:
			fprintf(stderr, "Unknown budgeting ignored\n");
			return;
	}

	/* only DepQBF budgeting leaves the switch, after setting budget */
	mprintf4(("%d: setting budget via %s\n", 
		  cmdq.rank, budget));
	ret = qdpll_configure(cmdq.solver, budget);
	free(budget);
	if (ret != NULL)
		fprintf(stderr,"%d: error message %s\n",
			cmdq.rank, ret);
	return;
}

void set_assumptions(const long *lits, const int nlits)
{
	int i, var, lit;
	if (!cmdq.is_sat) /* using DepQBF */
		for (i=0; i<nlits; i++)
			qdpll_assume(cmdq.solver, lits[i]);
	else /* using Glucose */
		for (i=0; i<nlits; i++)
		{
			var = lits[i]>0 ? lits[i] : -1 * lits[i];
			var--;
			lit = lits[i]>0 ? minisat_mkLit(var) : 
					  minisat_negate(minisat_mkLit(var));
			minisat_solve_addLit(cmdq.sat_solver, lit);
		}
}

int run_solver(void)
{
	QDPLLResult res;
	minisat_lbool sres;
	if (!cmdq.is_sat) /* using DepQBF */
	{
		res = qdpll_sat(cmdq.solver);
		if (res == QDPLL_RESULT_UNSAT)
			return -1;
		else if (res == QDPLL_RESULT_SAT)
			return 1;
		else
			return 0;
	}
	/* using Glucose */

	sres = minisat_limited_solve_commit(cmdq.sat_solver);
	if (sres == minisat_l_False)
		return -1;
	else if (sres == minisat_l_True)
		return 1;
	else
		return 0;
}
